#
# Copyright 1996,1997, by Perry A. Stoll <pas@xis.xerox.com>.
# All rights reserved. 
#
# Permission to use, copy, modify, and distribute this
# software and its documentation for any purpose and without
# fee is hereby granted, provided that the above copyright
# notice appear in all copies and that both that copyright
# notice and this permission notice appear in supporting
# documentation, and that the name of the author not be used
# in advertising or publicity pertaining to distribution of
# the software without specific, written prior permission. 
#
#

"""Provide an interface between the Python interpreter opcode tracker
object and the pstats profile object.

See the opstrackerscript.py for a full demonstration.

Example usage:

>>> import optracker,opstats,pstats
>>> optr = optracker.CreateTracker()

>>> optracker.StartTracking(optr)
  ... run your application, e.g.
>>> import profile
# run some program, e.g., Python-1.4/Tools/scripts/pystone.py
>>> profile.run('import pystone; pystone.main()')
>>> optracker.StopTracking()

# print out the results...
>>> o = opstats.OpCodeStats(optr)
>>> p = opstats.OStats(o)

# now p is a normal pstat Stats object. See profile.doc for more usage
>>> p.sort_stats('time').print_stats()

# OStats object has a method to return the matrix of transition probabilities
>>> transition_probability_lists = o.calculate_prev_prob()

# if you have the excellent NumPy numeric package:
>>> from Numeric import *
>>> trans = array(transition_probability_lists)

"""
import optracker

_opcode_dict = optracker.opcode_dict

# make a sorted copy of the builtin opcode list
_opcode_sorted_list = _opcode_dict.keys()[:]
_opcode_sorted_list.sort()

class OpCodeStats:
     def __init__(self,optr):
	  self.stats = None
	  self.optr  = optr
     def create_stats(self):
	  global _opcode_sorted_list,_opcode_dict
	  stats = {}
	  for opcode in _opcode_sorted_list:
	       cnt = self.optr.getcount(opcode)
	       if (cnt != 0):
		    tt = self.optr.gettime(opcode) # total time
		    name = _opcode_dict[opcode]	   # opcode name
		    pc = tt/cnt		           # avg time of execution
		    stats[('opcode',opcode,name)] = (cnt,cnt,tt,tt,{})
	  self.stats = stats

     def calculate_prev_prob(self):
	  global _opcode_sorted_list

	  oplistsize = optracker.opcode_listsize
	  prob = [None]*oplistsize 
	  for opcode in range(oplistsize):
	       if (opcode in _opcode_sorted_list):
		    cnt = self.optr.getcount(opcode)
		    if cnt == 0:
			 prob[opcode] = [0.0]*oplistsize
		    else:
			 prev = self.optr.getprev(opcode)
			 cnt = float(cnt)
			 prob[opcode] = map(lambda x,y=cnt: x/y, prev)
	       else:
		    prob[opcode] = [0.0]*oplistsize
	  self.prob = prob
	  return prob

import string,pstats,fpformat
class OStats(pstats.Stats):
    
    def print_title(self):
	print string.rjust('ncalls', 9),
	print string.rjust('tottime', 8),
	print string.rjust('percall', 8),
	print string.rjust('cumtime', 8),
	print string.rjust('percall', 8),
	print 'opcode:opcode id:(NAME)'
	
    def print_line(self, func):  # hack : should print percentages
	cc, nc, tt, ct, callers = self.stats[func]
	c = `nc`
	if nc != cc:
	    c = c + '/' + `cc`
	print string.rjust(c, 9),
	print f8(tt),
	if nc == 0:
	    print ' '*8,
	else:
	    print f8(tt/nc,digs=6),
	print f8(ct),
	if cc == 0:
	    print ' '*8,
	else:
	    print f8(ct/cc,digs=6),
	print pstats.func_std_string(func)
    

#**************************************************************************
# The following functions support printing of reports
#**************************************************************************
# define our own f8 function, because our times are much
# smaller than pstats expected.
def f8(x,digs=3):
    return string.rjust(fpformat.fix(x, digs), 8)

