/* opcodetrackerobject.c
 *
 * Module implementing object and methods for tracking opcode
 * execution in the Python bytecode interepter.
 *
 * created  January 1996, for Python 1.3
 * modified January 1997, for Python 1.4
 *
 * Copyright 1996,1997, by Perry A. Stoll <pas@xis.xerox.com>
 * All rights reserved.
 * 
 * 
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that copyright notice and this permission
 * notice appear in supporting documentation, and that the name of the
 * author not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior
 * permission.
 *
 * THERE ARE NO WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT
 * SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
 * DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA
 * OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
*/

/* Opcodetracker objects */
#include "Python.h"
#include "opcode.h"

#include "optracker.h"


/* try to use gettimeofday if we have it. */
#if defined(HAVE_GETTIMEOFDAY)

#ifdef GETTIMEOFDAY_NO_TZ
#  define OPTR_GETTIME(x) gettimeofday(x)
#else
#  define OPTR_GETTIME(x) gettimeofday(x,(void *)NULL)
#endif

#define OPTR_ACCUM_TIME_DIFF(x,y,z) \
   (x).tv_sec  += ((y).tv_sec  - (z).tv_sec);  \
   (x).tv_usec += ((y).tv_usec - (z).tv_usec)
#define OPTR_ZERO_OUT(x) \
   (x).tv_sec = (x).tv_usec = 0
#define OPTR_TIME_IN_SEC(x) \
   (((x).tv_sec) + (((x).tv_usec) / 1000000.0))
#define OPTR_COPY_TIME(d,s) \
   (d).tv_sec = (s).tv_sec; \
   (d).tv_usec = (s).tv_usec
#else
#error Need gettimeofday()
#endif /* USE_TIME

/* If non-NULL, points to the current opcode profiling object */
PyObject *PyOpTracker_Object = NULL;

#define RETURN_NONE   Py_INCREF(Py_None); return Py_None
#define TRY(E) if(!(E)) return NULL
#define ASSERT(E,M) if(!(E)) return ErrorReturn(M)

#define CHECK_OPCODE(func,op,val) \
   if (((op) < 0) || ((op) >= OPTRACK_LIST_SIZE)){ \
        fprintf(stderr,"%s: invalid opcode %d\n",(func),(op)); \
        return val; \
   } 
    
static PyObject *PyOpTracker_New Py_PROTO((void));
static void PyOpTracker_Dealloc Py_PROTO((PyOpTrackerObject *op));
static void PyOpTracker_Clear   Py_PROTO((PyOpTrackerObject *op));
static long PyOpTracker_GetCount Py_PROTO((PyOpTrackerObject *op, int opcode));
static double PyOpTracker_GetTime Py_PROTO((PyOpTrackerObject *op, int opcode));
static void PyOpTracker_Executed Py_PROTO((PyOpTrackerObject *op, int opcode));
static void PyOpTracker_Executed_Prev Py_PROTO((PyOpTrackerObject *op, int opcode,int prev_op));


/* array to hold names of opcodes */
static char opcode_names[OPTRACK_LIST_SIZE][OPTRACK_MAXNAME_SIZE] =
{
    /* generated the following by parsing the file Include/opcode.h
       and then hand modified to add the 'Also uses XXX-YYY' in SLICE,
       STORE_SLICE,DELETE_SLICE */

    "STOP_CODE",                /* opcode 0 */ 
    "POP_TOP",                  /* opcode 1 */ 
    "ROT_TWO",                  /* opcode 2 */ 
    "ROT_THREE",                /* opcode 3 */ 
    "DUP_TOP",                  /* opcode 4 */ 
    "",           /* non-opcode 5 */
    "",           /* non-opcode 6 */
    "",           /* non-opcode 7 */
    "",           /* non-opcode 8 */
    "",           /* non-opcode 9 */
    "UNARY_POSITIVE",           /* opcode 10 */ 
    "UNARY_NEGATIVE",           /* opcode 11 */ 
    "UNARY_NOT",                /* opcode 12 */ 
    "UNARY_CONVERT",            /* opcode 13 */ 
    "",           /* non-opcode 14 */
    "UNARY_INVERT",             /* opcode 15 */ 
    "",           /* non-opcode 16 */
    "",           /* non-opcode 17 */
    "",           /* non-opcode 18 */
    "BINARY_POWER",             /* opcode 19 */ 
    "BINARY_MULTIPLY",          /* opcode 20 */ 
    "BINARY_DIVIDE",            /* opcode 21 */ 
    "BINARY_MODULO",            /* opcode 22 */ 
    "BINARY_ADD",               /* opcode 23 */ 
    "BINARY_SUBTRACT",          /* opcode 24 */ 
    "BINARY_SUBSCR",            /* opcode 25 */ 
    "",           /* non-opcode 26 */
    "",           /* non-opcode 27 */
    "",           /* non-opcode 28 */
    "",           /* non-opcode 29 */
    "SLICE",                    /* opcode 30 */ 
    "SLICE1",                   /* opcode 31 */ 
    "SLICE2",                   /* opcode 32 */ 
    "SLICE3",                   /* opcode 33 */ 
    "",           /* non-opcode 34 */
    "",           /* non-opcode 35 */
    "",           /* non-opcode 36 */
    "",           /* non-opcode 37 */
    "",           /* non-opcode 38 */
    "",           /* non-opcode 39 */
    "STORE_SLICE",              /* opcode 40 */ 
    "STORE_SLICE1",             /* opcode 41 */ 
    "STORE_SLICE2",             /* opcode 42 */ 
    "STORE_SLICE3",             /* opcode 43 */ 
    "",           /* non-opcode 44 */
    "",           /* non-opcode 45 */
    "",           /* non-opcode 46 */
    "",           /* non-opcode 47 */
    "",           /* non-opcode 48 */
    "",           /* non-opcode 49 */
    "DELETE_SLICE",             /* opcode 50 */ 
    "DELETE_SLICE1",            /* opcode 51 */ 
    "DELETE_SLICE2",            /* opcode 52 */ 
    "DELETE_SLICE3",            /* opcode 53 */ 
    "",           /* non-opcode 54 */
    "",           /* non-opcode 55 */
    "",           /* non-opcode 56 */
    "",           /* non-opcode 57 */
    "",           /* non-opcode 58 */
    "",           /* non-opcode 59 */
    "STORE_SUBSCR",             /* opcode 60 */ 
    "DELETE_SUBSCR",            /* opcode 61 */ 
    "BINARY_LSHIFT",            /* opcode 62 */ 
    "BINARY_RSHIFT",            /* opcode 63 */ 
    "BINARY_AND",               /* opcode 64 */ 
    "BINARY_XOR",               /* opcode 65 */ 
    "BINARY_OR",                /* opcode 66 */ 
    "",           /* non-opcode 67 */
    "",           /* non-opcode 68 */
    "",           /* non-opcode 69 */
    "PRINT_EXPR",               /* opcode 70 */ 
    "PRINT_ITEM",               /* opcode 71 */ 
    "PRINT_NEWLINE",            /* opcode 72 */ 
    "",           /* non-opcode 73 */
    "",           /* non-opcode 74 */
    "",           /* non-opcode 75 */
    "",           /* non-opcode 76 */
    "",           /* non-opcode 77 */
    "",           /* non-opcode 78 */
    "",           /* non-opcode 79 */
    "BREAK_LOOP",               /* opcode 80 */ 
    "",           /* non-opcode 81 */
    "LOAD_LOCALS",              /* opcode 82 */ 
    "RETURN_VALUE",             /* opcode 83 */ 
    "",           /* non-opcode 84 */
    "EXEC_STMT",                /* opcode 85 */ 
    "",           /* non-opcode 86 */
    "POP_BLOCK",                /* opcode 87 */ 
    "END_FINALLY",              /* opcode 88 */ 
    "BUILD_CLASS",              /* opcode 89 */ 
    "STORE_NAME",               /* opcode 90 */ 
    "DELETE_NAME",              /* opcode 91 */ 
    "UNPACK_TUPLE",             /* opcode 92 */ 
    "UNPACK_LIST",              /* opcode 93 */ 
    "UNPACK_ARG",               /* opcode 94 */ 
    "STORE_ATTR",               /* opcode 95 */ 
    "DELETE_ATTR",              /* opcode 96 */ 
    "STORE_GLOBAL",             /* opcode 97 */ 
    "DELETE_GLOBAL",            /* opcode 98 */ 
    "UNPACK_VARARG",            /* opcode 99 */ 
    "LOAD_CONST",               /* opcode 100 */ 
    "LOAD_NAME",                /* opcode 101 */ 
    "BUILD_TUPLE",              /* opcode 102 */ 
    "BUILD_LIST",               /* opcode 103 */ 
    "BUILD_MAP",                /* opcode 104 */ 
    "LOAD_ATTR",                /* opcode 105 */ 
    "COMPARE_OP",               /* opcode 106 */ 
    "IMPORT_NAME",              /* opcode 107 */ 
    "IMPORT_FROM",              /* opcode 108 */ 
    "ACCESS_MODE",              /* opcode 109 */ 
    "JUMP_FORWARD",             /* opcode 110 */ 
    "JUMP_IF_FALSE",            /* opcode 111 */ 
    "JUMP_IF_TRUE",             /* opcode 112 */ 
    "JUMP_ABSOLUTE",            /* opcode 113 */ 
    "FOR_LOOP",                 /* opcode 114 */ 
    "LOAD_LOCAL",               /* opcode 115 */ 
    "LOAD_GLOBAL",              /* opcode 116 */ 
    "SET_FUNC_ARGS",            /* opcode 117 */ 
    "",           /* non-opcode 118 */
    "",           /* non-opcode 119 */
    "SETUP_LOOP",               /* opcode 120 */ 
    "SETUP_EXCEPT",             /* opcode 121 */ 
    "SETUP_FINALLY",            /* opcode 122 */ 
    "",           /* non-opcode 123 */
    "LOAD_FAST",                /* opcode 124 */ 
    "STORE_FAST",               /* opcode 125 */ 
    "DELETE_FAST",              /* opcode 126 */ 
    "SET_LINENO",               /* opcode 127 */ 
    "",           /* non-opcode 128 */
    "",           /* non-opcode 129 */
    "RAISE_VARARGS",            /* opcode 130 */ 
    "CALL_FUNCTION",            /* opcode 131 */ 
    "MAKE_FUNCTION",            /* opcode 132 */
    "BUILD_SLICE",              /* opcode 133 */
};

/* some utility routines */

static PyObject *ErrorObject;

static PyObject *
ErrorReturn(char *message)
{
  PyErr_SetString(ErrorObject,message);
  return NULL;
}

static int
optr_valid_opcode(int opcode)
{
    if ((opcode < 0) || (opcode >= OPTRACK_LIST_SIZE)){
        char msg[128];
        sprintf(msg,"opcode must be between 0 and %d",OPTRACK_LIST_SIZE-1);
        PyErr_SetString(PyExc_TypeError,msg);
        return 0;
    }
    return 1;
}

static void
optr_init_opcode(OpCodeTracker_Info *arg)
{
    int ii;
    for(ii = 0; ii < OPTRACK_LIST_SIZE; ii++){
        arg->op_prev[ii] = 0;
    }
    arg->op_cnt = 0;
    OPTR_ZERO_OUT(arg->op_cum_ticks);
}


PyObject *
optr_make_opcode_dict(void)
{
    PyObject *thedict,*theint,*thestr;
    int ii;
   
    TRY(thedict = PyDict_New());
    for(ii = 0; ii < OPTRACK_LIST_SIZE; ii++){
        if (strcmp(opcode_names[ii],"")){
            theint = PyInt_FromLong(ii);
            thestr = PyString_FromString(opcode_names[ii]);
            
            /* dict does not take over ownership so DECREF args.*/
            if ((theint==NULL) || (thestr==NULL)
                || (PyDict_SetItem(thedict,theint,thestr)< 0)){
                Py_XDECREF(theint);
                Py_XDECREF(thestr);
                Py_DECREF(thedict);
                return NULL;
            }
            Py_DECREF(theint);
            Py_DECREF(thestr);
        }
    }
    return thedict;
}

static PyObject *
optr_make_list_from_longints(long int *items, int len)
{
    PyObject *thelist,*theint;
    int ii;

    TRY(thelist = PyList_New(len));

    for(ii = 0; ii < len; ii++){
        theint = PyInt_FromLong(items[ii]);
        /* list takes over ownership of theint, so don't DECREF it.*/
        if ((theint == NULL) || (PyList_SetItem(thelist,ii,theint) < 0)){
            Py_DECREF(thelist);
            return NULL;
        }
    }
    return thelist;
}

/* Python C interface functions for manipulating OpTrackers: */

PyObject *
PyOpTracker_New(void)
{
    register PyOpTrackerObject *op;
    
    TRY(op = PyObject_NEW(PyOpTrackerObject, &PyOpTracker_Type));
    PyOpTracker_Clear(op);
    op->prev_op = op->timing_op = NOOPCODE;
    return (PyObject *)op;
}

  
void
PyOpTracker_Clear(PyOpTrackerObject *op)
{
    int ii;

    for(ii = 0; ii < OPTRACK_LIST_SIZE; ii++){
        optr_init_opcode(op->ops+ii);
    }
}

void
PyOpTracker_Install(PyObject *op)
{
    Py_XINCREF(op);
    Py_XDECREF(PyOpTracker_Object);
    PyOpTracker_Object = op;
}

long
PyOpTracker_GetCount(PyOpTrackerObject *op, int opcode)
{
    CHECK_OPCODE("PyOpTracker_GetCount",opcode,0);
    return op->ops[opcode].op_cnt;
}

static double
PyOpTracker_GetTime(PyOpTrackerObject *op, int opcode)
{
    CHECK_OPCODE("PyOpTracker_GetTime",opcode,0.0);
    return OPTR_TIME_IN_SEC(op->ops[opcode].op_cum_ticks);
}

void
PyOpTracker_StartOpTime(PyOpTrackerObject *op, int opcode)
{
    CHECK_OPCODE("PyOpTracker_StartOpTime",opcode,);
    if (op->timing_op != NOOPCODE){
        PyOpTracker_EndOpTime(op,op->timing_op);
    }
    op->ops[opcode].op_cnt++;
    if (op->prev_op != NOOPCODE){
        op->ops[opcode].op_prev[op->prev_op]++;
    }
    op->timing_op = opcode;
    OPTR_GETTIME(&op->start);
}

void
PyOpTracker_EndOpTime(PyOpTrackerObject *op, int opcode)
{
    struct timeval end; 
    
    CHECK_OPCODE("PyOpTracker_EndOpTime",opcode,);
    if (op->timing_op == NOOPCODE){
        fprintf(stderr,"EndOpTime: not timing. prev_op=%d opcode=%d\n",
                op->prev_op,opcode);
        return;
    }
    OPTR_GETTIME(&end);
    OPTR_ACCUM_TIME_DIFF(op->ops[op->timing_op].op_cum_ticks,end,op->start);
    OPTR_COPY_TIME(op->start,end);
    op->prev_op = op->timing_op;
    op->timing_op = NOOPCODE;
}


/* optracker methods */

void
PyOpTracker_Dealloc(PyOpTrackerObject *op)
{
    PyMem_DEL(op);
}

static char optracker_clear_one_doc[] = "Set all stats to initial values (generally, zero) for the given OPCODE.";

static PyObject *
optracker_clear_one(PyOpTrackerObject *self,PyObject *args)
{
    int opcode;
    
    TRY(PyArg_ParseTuple(args, "i", &opcode));
    TRY(optr_valid_opcode(opcode));
    optr_init_opcode(self->ops+opcode);
    RETURN_NONE;
}

static char optracker_clear_all_doc[] = "Set all stats to initial values (generally, zero) for all opcodes.";

static PyObject *
optracker_clear_all(PyOpTrackerObject *self,PyObject *args)
{
    TRY(PyArg_ParseTuple(args, ""));
    PyOpTracker_Clear(self);
    RETURN_NONE;
}

static char optracker_getcount_doc[] = "Returns the number of times the given OPCODE has been executed.";

static PyObject *
optracker_getcount(PyOpTrackerObject *self,PyObject *args)
{
    int opcode;
    int opcount;
    
    TRY(PyArg_ParseTuple(args, "i", &opcode));
    TRY(optr_valid_opcode(opcode));
    return Py_BuildValue("l", PyOpTracker_GetCount(self,opcode));
}

static char optracker_gettime_doc[] = "Returns the cumulative time in seconds the given OPCODE has executed.";


static PyObject *
optracker_gettime(PyOpTrackerObject *self,PyObject *args)
{
    int opcode;
    PyOpTrackerObject *op;
    
    TRY(PyArg_ParseTuple(args, "i", &opcode));
    TRY(optr_valid_opcode(opcode));
    return Py_BuildValue("d", PyOpTracker_GetTime(self,opcode));
}


static char optracker_getname_doc[] = "Returns the name of the specified OPCODE.";


static PyObject *
optracker_getname(PyOpTrackerObject *self,PyObject *args)
{
    int opcode;
    double optime;
    PyOpTrackerObject *op;
    
    TRY(PyArg_ParseTuple(args, "i", &opcode));
    TRY(optr_valid_opcode(opcode));
    return PyString_FromString(opcode_names[opcode]);
}

static char optracker_getprev_doc[] = "Given an OPCODE, returns a list showing how often that OPCODE was called following the other opcodes.";

static PyObject *
optracker_getprev(PyOpTrackerObject *self,PyObject *args)
{
    int opcode;
    PyObject *list;
    PyOpTrackerObject *op;
    
    TRY(PyArg_ParseTuple(args, "i", &opcode));
    TRY(optr_valid_opcode(opcode));
    TRY(list =
        optr_make_list_from_longints(self->ops[opcode].op_prev,OPTRACK_LIST_SIZE));
    return list;
}

struct PyMethodDef optracker_methods[] = {
    {"clearall",   (PyCFunction)optracker_clear_all,1,optracker_clear_all_doc},
    {"clearone",   (PyCFunction)optracker_clear_one,1,optracker_clear_one_doc},
    {"getcount",(PyCFunction)optracker_getcount,1,optracker_getcount_doc},
    {"gettime", (PyCFunction)optracker_gettime,1,optracker_gettime_doc},
    {"getname", (PyCFunction)optracker_getname,1,optracker_getname_doc},
    {"getprev", (PyCFunction)optracker_getprev,1,optracker_getprev_doc},
    {NULL,		NULL, }		/* sentinel */
};


static char PyOpTracker_Type_doc[] = "Type used in tracking execution of opcodes in the Python interpreter, including execution count, time, and sequencing of each opcode.";

static PyObject *
optracker_repr(PyOpTrackerObject *self)
{
    PyObject *ret;
    char string[64];

    sprintf(string,"<optracker at 0x%x>",self);
    return PyString_FromString(string);
}

static PyObject *
optracker_getattr(PyOpTrackerObject *self, char * name)
{
    if (!strcmp(name,"__doc__"))
        return PyString_FromString(PyOpTracker_Type_doc);
    return Py_FindMethod(optracker_methods, (PyObject *)self, name);
}

PyTypeObject PyOpTracker_Type = {
	PyObject_HEAD_INIT(&PyType_Type)
	0,			/*ob_size*/
	"optrackerobject",	/*tp_name*/
	sizeof(PyOpTrackerObject),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)PyOpTracker_Dealloc, /*tp_dealloc*/
	0,			/*tp_print*/
	(getattrfunc)optracker_getattr, /*tp_getattr*/
	0,                      /*tp_setattr*/
	0,			/*tp_compare*/
	(reprfunc)optracker_repr,/*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
        (ternaryfunc)0,		/*tp_call*/
	(reprfunc)0,		/*tp_str*/

	/* Space for future expansion */
	0L,0L,0L,0L,
	PyOpTracker_Type_doc   /* Documentation string */
};


static char OpTracker_CreateTracker__doc__[] =
"Returns a new opcode tracker object used to profile the Python interpreter.";


static PyObject *
OpTracker_CreateTracker(PyObject *notused, PyObject *args)
{
    TRY(PyArg_ParseTuple(args, ""));
    return PyOpTracker_New();
}

static char OpTracker_StartTracker__doc__[] =
"Installs the opcode tracker object so that it will be used by the Python interpreter to track opcode execution.";

static PyObject *
OpTracker_StartTracker(PyObject *notused, PyObject *args)
{
    PyObject *op;
    PyObject *globals;
    
    TRY(PyArg_ParseTuple(args, "O!",&PyOpTracker_Type,&op));
    PyOpTracker_Install(op);

    /* ToDo: install object in global dictionary. This way the
       interpeter can retreive this object without being statically
       linked. Too slow? */
    /* Try to put the object in the global dictionary. Don't worry if
       we fail. */
    globals = PyEval_GetGlobals();
    if (globals){
        PyDict_SetItemString(globals,OPTRACK_OBJ_STRING,op);
        PyErr_Clear();
    }
    RETURN_NONE;
}


static char OpTracker_StopTracker__doc__[] =
"Removes any opcode tracker object from the Python interpreter.";


static PyObject *
OpTracker_StopTracker(PyObject *notused, PyObject *args)
{
    PyObject *op;
    PyObject *globals;
    
    TRY(PyArg_ParseTuple(args, ""));
    PyOpTracker_Install(NULL);

    /* ToDo: install object in global dictionary. This way the
       interpeter can retreive this object without being statically
       linked. Too slow? */
    globals = PyEval_GetGlobals();
    if (globals){
        PyDict_DelItemString(globals,OPTRACK_OBJ_STRING);
        PyErr_Clear();
    }
    RETURN_NONE;
}

/* List of methods defined in the module */

static struct PyMethodDef OpTracker_methods[] = {
    {"CreateTracker",OpTracker_CreateTracker,1,OpTracker_CreateTracker__doc__},
    {"StartTracking",OpTracker_StartTracker,1,OpTracker_StartTracker__doc__},
    {"StopTracking",OpTracker_StopTracker,1,OpTracker_StopTracker__doc__},
    {NULL, NULL} /* sentinel */
};

/* Initialization function for the module (*must* be called initoptracker) */

static char OpTracker_module_documentation[] = 
"optracker provides a way to collect execution stats of the Python interpreter by keeping track of frequency,time, and history of opcode execution.";

void
initoptracker()
{
    PyObject *m, *d, *o;
    
    /* Create the module and add the functions */
    m = Py_InitModule4("optracker", OpTracker_methods,
                       OpTracker_module_documentation,
                       (PyObject*)NULL,PYTHON_API_VERSION);
    
    /* Add some symbolic constants to the module */
    d = PyModule_GetDict(m);
    ErrorObject = PyString_FromString("OptrackerError");
    PyDict_SetItemString(d, "error", ErrorObject);

    o = optr_make_opcode_dict();
    PyDict_SetItemString(d, "opcode_dict", o);

    o = PyInt_FromLong(OPTRACK_LIST_SIZE);
    PyDict_SetItemString(d, "opcode_listsize", o);
    
    /* Check for errors */
    if (PyErr_Occurred())
        Py_FatalError("can't initialize module optracker");
}

